from __future__ import annotations

import configparser
import logging
from typing import TYPE_CHECKING, Any, Iterable, List, Mapping, Optional, Type

from configparser_override._strategy_factory import StrategyFactory
from configparser_override.convert import ConfigConverter

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

    from configparser_override._override_strategy import Strategy
    from configparser_override.types import Dataclass, _optionxform_fn

logger = logging.getLogger(__name__)


class ConfigParserOverride:
    def __init__(
        self,
        env_prefix: str = "",
        create_new_from_env_prefix: bool = True,
        create_new_from_direct: bool = True,
        config_parser: configparser.ConfigParser | None = None,
        case_sensitive_overrides: bool = False,
        optionxform: _optionxform_fn | None = None,
        **overrides: str,
    ):
        """
        Initialize the ConfigParserOverride.

        :param env_prefix: Optional prefix for environment variables,
            defaults to an empty string.
        :type env_prefix: str, optional
        :param create_new_from_env_prefix: Flag to create new configuration
            options from environment variables.
        :type create_new_from_env_prefix: bool, optional
        :param create_new_from_direct: Flag to create new configuration
            options from direct overrides.
        :type create_new_from_direct: bool, optional
        :param config_parser: Optional ConfigParser object to be used,
            defaults to None.
        :type config_parser: configparser.ConfigParser, optional
        :param case_sensitive_overrides: Flag to indicate if overrides should
            be case sensitive.
        :type case_sensitive_overrides: bool, optional
        :param optionxform: Optional function to transform option strings.
        :type optionxform: _optionxform_fn | None, optional
        :param overrides: Keyword arguments to directly override configuration values.
        :type overrides: dict[str, str | None]

        **Examples:**

        .. code-block:: python

            >>> parser_override = ConfigParserOverride(env_prefix='MYAPP_', test_option='value')
            >>> parser_override.read(['example.ini'])
            >>> parser_override.apply_overrides()
            >>> config = parser_override.config
            >>> config.get('DEFAULT', 'test_option')
            'value'
        """

        self.env_prefix = env_prefix
        self.create_new_from_env_prefix = create_new_from_env_prefix
        self.create_new_from_direct = create_new_from_direct
        self.case_sensitive_overrides = case_sensitive_overrides
        self.optionxform = optionxform
        self.overrides = overrides

        # Configure ConfigParser and align optionxform for consistency in later
        # inferance for overrides
        if config_parser is None:
            self._config = configparser.ConfigParser()
            if self.optionxform is not None:
                self._config.optionxform = self.optionxform  # type: ignore
        else:
            self._config = config_parser
            self.optionxform = self._config.optionxform

    def _get_override_strategy(self) -> Strategy:
        """
        Get the appropriate override strategy based on initialization parameters.

        :return: The appropriate strategy instance.
        :rtype: Strategy
        """
        return StrategyFactory(
            self._config,
            self.env_prefix,
            self.create_new_from_env_prefix,
            self.create_new_from_direct,
            self.overrides,
            self.case_sensitive_overrides,
            self.optionxform,
        ).get_strategy()

    def apply_overrides(self) -> None:
        """
        Apply overrides to the current configuration.

        This method utilizes the override strategy based on initialization parameters
        to apply the overrides to the current configuration.

        **Examples:**

        .. code-block:: python

            >>> parser_override = ConfigParserOverride(test_option='value')
            >>> parser_override.read(['example.ini'])
            >>> parser_override.apply_overrides()
            >>> config = parser_override.config
            >>> config.get('DEFAULT', 'test_option')
            'value'
        """
        strategy = self._get_override_strategy()
        strategy.execute()

    def read_dict(
        self, dictionary: Mapping[str, Mapping[str, Any]], source: str = "<string>"
    ) -> None:
        """
        Read configuration from a dictionary.

        This method reads the configuration data from a dictionary.

        :param dictionary: The dictionary containing configuration data.
        :type dictionary: Mapping[str, Mapping[str, Any]]
        :param source: The source name of the dictionary being read, defaults to "<string>".
        :type source: str, optional

        **Examples:**

        .. code-block:: python

            >>> parser_override = ConfigParserOverride()
            >>> config_dict = {'section1': {'key1': 'value1'}}
            >>> parser_override.read_dict(config_dict)
            >>> parser_override.apply_overrides()
            >>> config = parser_override.config
            >>> config.get('section1', 'key1')
            'value1'
        """
        self._config.read_dict(dictionary, source=source)

    def read_string(self, string: str, source: str = "<string>") -> None:
        """
        Read configuration from a string.

        This method reads the configuration data from a string.

        :param string: The string containing configuration data.
        :type string: str
        :param source: The source name of the string being read, defaults to "<string>".
        :type source: str, optional

        **Examples:**

        .. code-block:: python

            >>> parser_override = ConfigParserOverride()
            >>> config_string = \"\"\"
            ... [section1]
            ... key1 = value1
            ... \"\"\"
            >>> parser_override.read_string(config_string)
            >>> parser_override.apply_overrides()
            >>> config = parser_override.config
            >>> config.get('section1', 'key1')
            'value1'
        """
        self._config.read_string(string, source=source)

    def read_file(self, f: Iterable[str], source: str | None = None) -> None:
        """
        Read configuration from a file-like object.

        This method reads the configuration data from a file-like object.

        :param f: An iterable of strings representing lines in a file-like object.
        :type f: Iterable[str]
        :param source: The source name of the file being read, defaults to None.
        :type source: str | None, optional

        **Examples:**

        .. code-block:: python

            >>> from io import StringIO
            >>> parser_override = ConfigParserOverride()
            >>> file_content = StringIO(\"\"\"
            ... [section1]
            ... key1 = value1
            ... \"\"\")
            >>> parser_override.read_file(file_content)
            >>> parser_override.apply_overrides()
            >>> config = parser_override.config
            >>> config.get('section1', 'key1')
            'value1'
        """
        self._config.read_file(f, source=source)

    def read(
        self,
        filenames: StrOrBytesPath | Iterable[StrOrBytesPath],
        encoding: str | None = None,
    ) -> list[str]:
        """
        Read configuration from one or more files.

        This method is a wrapper around :py:meth:`configparser.ConfigParser.read` that
        reads the specified filenames in order.

        :param filenames: A single filename or an iterable of filenames to read.
        :type filenames: :py:class:`_typeshed.StrOrBytesPath` or
            Iterable[:py:class:`_typeshed.StrOrBytesPath`]
        :param encoding: The encoding to use for reading the files, defaults to None.
        :type encoding: str, optional
        :return: List of successfully parsed files
        :rtype: list[str]

        **Examples:**

        .. code-block:: python

            >>> parser_override = ConfigParserOverride(test_option='value')
            >>> parser_override.read(['example.ini'])
            >>> parser_override.apply_overrides()
            >>> config = parser_override.config
            >>> config.get('DEFAULT', 'test_option')
            'value'
        """
        files_read = self._config.read(filenames=filenames, encoding=encoding)
        return files_read

    @property
    def config(self) -> configparser.ConfigParser:
        """
        Property to access the configuration.

        This can be used to modify the property of the configparser object and
        also set and get options manually.

        :return: The :py:class:`configparser.ConfigParser` object
            with the configuration.
        :rtype: :py:class:`configparser.ConfigParser`

        **Examples:**

        Get an option after parsing and overrides:

        .. code-block:: python

            >>> config = ConfigParserOverride(test_option='value')
            >>> config.read(['example.ini'])
            >>> config.apply_overrides()
            >>> config.get('DEFAULT', 'test_option')
            'value'

        Can also be used like just like regular ConfigParser:

        .. code-block:: python

            >>> parser_override = ConfigParserOverride()
            >>> config = parser_override.config
            >>> config.set('section', 'option', 'value')
            >>> config.get('section', 'option')
            'value'
        """
        return self._config

    def to_dataclass(
        self,
        dataclass: Type[Dataclass],
        include_sections: Optional[List[str]] = None,
        exclude_sections: Optional[List[str]] = None,
    ) -> Dataclass:
        """
        Convert the configuration data to a dataclass instance.

        This method allows converting the configuration into a dataclass instance,
        optionally including or excluding specific sections. It can be useful for
        validating that the configuration adheres to the expected format and leveraging
        various typing frameworks, e.g., integrations with text editors.

        :param dataclass: The dataclass type to convert the configuration data into.
        :type dataclass: Dataclass
        :param include_sections: A list of section names to explicitly include in the
                                 conversion. If provided, only these sections will be
                                 included in the resulting dataclass.
                                 Default is None.
        :type include_sections: Optional[List[str]]
        :param exclude_sections: A list of section names to exclude from the conversion.
                                 If provided, these sections will be excluded from the
                                 resulting dataclass.
                                 Default is None.
        :type exclude_sections: Optional[List[str]]
        :return: An instance of the dataclass populated with the configuration data.
        :rtype: Dataclass

        **Examples:**

        .. code-block:: python

            >>> from dataclasses import dataclass

            >>> @dataclass
            ... class Section1:
            ...     key: str

            >>> @dataclass
            ... class ExampleConfig:
            ...     section1: Section1

            >>> config_parser_override = ConfigParserOverride(section1__key="a string")
            >>> config_parser_override.read()
            >>> config_parser_override.apply_overrides()
            >>> config_as_dataclass = config_parser_override.to_dataclass(ExampleConfig)
            >>> assert config_as_dataclass.section1.key == "a string" # True
        """

        return ConfigConverter(
            config=self._config,
            include_sections=include_sections,
            exclude_sections=exclude_sections,
        ).to_dataclass(dataclass)



from typing import TYPE_CHECKING, Protocol, TypeVar

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

    Dataclass = TypeVar("Dataclass", bound=DataclassInstance)


class _optionxform_fn(Protocol):
    def __call__(self, optionstr: str) -> str: ...  # pragma: no cover



from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from configparser_override.exceptions import NoConfigFilesFoundError
from configparser_override.file_collector import (
    _log_and_return_if_exists,
    _unix_collect_home_config,
    _unix_collect_system_config,
    _windows_collect_home_config,
    _windows_collect_system_config,
    config_file_collector,
)


# Mock logger to avoid actual logging during tests
@pytest.fixture(autouse=True)
def mock_logger():
    with patch("configparser_override.file_collector.logger"):
        yield


def test_log_and_return_if_exists_file_exists():
    path = MagicMock(spec=Path)
    path.exists.return_value = True
    assert _log_and_return_if_exists(path) == path


def test_log_and_return_if_exists_file_does_not_exist():
    path = MagicMock(spec=Path)
    path.exists.return_value = False
    assert _log_and_return_if_exists(path) is None


@patch("configparser_override.file_collector.Path.home")
@patch("configparser_override.file_collector.os.getenv")
def test_unix_collect_home_config(mock_getenv, mock_home):
    mock_home.return_value = Path("/home/testuser")
    mock_getenv.return_value = Path("/home/testuser/.config")
    subdir = "testapp"
    file_name = "config.ini"

    with patch(
        "configparser_override.file_collector._log_and_return_if_exists"
    ) as mock_log_and_return_if_exists:
        mock_log_and_return_if_exists.return_value = Path(
            "/home/testuser/.config/testapp/config.ini"
        )
        result = _unix_collect_home_config(subdir, file_name)
        assert result == Path("/home/testuser/.config/testapp/config.ini")


@patch("configparser_override.file_collector.os.getenv")
@patch("configparser_override.file_collector.Path.exists")
def test_unix_collect_system_config(mock_path_exists, mock_os_getenv):
    mock_os_getenv.return_value = "/etc/xdg"

    mock_path_exists.return_value = True
    subdir = "testapp"
    file_name = "config.ini"
    result = _unix_collect_system_config(subdir, file_name)
    assert result == [Path("/etc/xdg/testapp/config.ini")]


@patch("configparser_override.file_collector.os.getenv")
@patch("configparser_override.file_collector.Path.exists")
def test_unix_collect_system_multi_config(mock_path_exists, mock_os_getenv):
    # First value is most important
    mock_os_getenv.return_value = "/etc/xdg:/etc"

    mock_path_exists.return_value = True
    subdir = "testapp"
    file_name = "config.ini"
    result = _unix_collect_system_config(subdir, file_name)
    # Reverse importance for configpraser override default
    assert result == [
        Path("/etc/testapp/config.ini"),
        Path("/etc/xdg/testapp/config.ini"),
    ]


@patch("configparser_override.file_collector.os.getenv")
@patch("configparser_override.file_collector.Path.exists")
def test_unix_collect_system_multi_config_not_exist(mock_path_exists, mock_os_getenv):
    # First value is most important
    mock_os_getenv.return_value = "/etc/xdg:/etc"

    mock_path_exists.return_value = False
    subdir = "testapp"
    file_name = "config.ini"
    result = _unix_collect_system_config(subdir, file_name)
    # Reverse importance for configpraser override default
    assert result == []


@patch("configparser_override.file_collector.Path.exists")
def test_unix_collect_system_bare_etc(mock_path_exists):
    mock_path_exists.return_value = True
    subdir = "testapp"
    file_name = "config.ini"
    result = _unix_collect_system_config(subdir, file_name, bare_etc=True)
    # Reverse importance for configpraser override default
    assert result == [
        Path("/etc/testapp/config.ini"),
    ]


@patch("configparser_override.file_collector.os.getenv")
@patch("configparser_override.file_collector.Path.exists")
def test_windows_collect_home_config(mock_path_exists, mock_os_getenv):
    mock_os_getenv.return_value = "C:/Users/testuser/AppData/Roaming"
    mock_path_exists.return_value = True
    subdir = "testapp"
    file_name = "config.ini"
    result = _windows_collect_home_config(subdir, file_name)
    assert result == Path("C:/Users/testuser/AppData/Roaming/testapp/config.ini")


@patch("configparser_override.file_collector.os.getenv")
def test_windows_collect_home_config_not_exist(mock_os_getenv):
    mock_os_getenv.return_value = None
    subdir = "testapp"
    file_name = "config.ini"
    result = _windows_collect_home_config(subdir, file_name)
    assert result is None


@patch("configparser_override.file_collector.os.getenv")
@patch("configparser_override.file_collector.Path.exists")
def test_windows_collect_system_config(mock_path_exists, mock_os_getenv):
    mock_os_getenv.return_value = "C:/ProgramData"
    mock_path_exists.return_value = True
    subdir = "testapp"
    file_name = "config.ini"
    result = _windows_collect_system_config(subdir, file_name)
    assert result == [Path("C:/ProgramData/testapp/config.ini")]


@patch("configparser_override.file_collector.os.getenv")
@patch("configparser_override.file_collector.Path.exists")
def test_windows_collect_system_config_not_exist_path(mock_path_exists, mock_os_getenv):
    mock_os_getenv.return_value = "C:/ProgramData"
    mock_path_exists.return_value = False
    subdir = "testapp"
    file_name = "config.ini"
    result = _windows_collect_system_config(subdir, file_name)
    assert result == []


@patch("configparser_override.file_collector.os.getenv")
def test_windows_collect_system_config_not_exist(mock_os_getenv):
    mock_os_getenv.return_value = None
    subdir = "testapp"
    file_name = "config.ini"
    result = _windows_collect_system_config(subdir, file_name)
    assert result == []


@patch("configparser_override.file_collector.platform.system", return_value="Windows")
@patch("configparser_override.file_collector._windows_collect_system_config")
@patch("configparser_override.file_collector._windows_collect_home_config")
def test_config_file_collector_windows(
    mock_home_config, mock_system_config, mock_platform
):
    mock_system_config.return_value = [Path("C:/ProgramData/testapp/config.ini")]
    mock_home_config.return_value = Path(
        "C:/Users/testuser/AppData/Roaming/testapp/config.ini"
    )
    result = config_file_collector("config.ini", "testapp")
    assert result == [
        Path("C:/ProgramData/testapp/config.ini"),
        Path("C:/Users/testuser/AppData/Roaming/testapp/config.ini"),
    ]


@patch("configparser_override.file_collector.platform.system", return_value="Linux")
@patch("configparser_override.file_collector._unix_collect_system_config")
@patch("configparser_override.file_collector._unix_collect_home_config")
def test_config_file_collector_linux(
    mock_home_config, mock_system_config, mock_platform
):
    mock_system_config.return_value = [Path("/etc/xdg/testapp/config.ini")]
    mock_home_config.return_value = Path("/home/testuser/.config/testapp/config.ini")
    result = config_file_collector("config.ini", "testapp")
    assert result == [
        Path("/etc/xdg/testapp/config.ini"),
        Path("/home/testuser/.config/testapp/config.ini"),
    ]


@patch("configparser_override.file_collector.platform.system", return_value="Darwin")
@patch("configparser_override.file_collector._unix_collect_system_config")
@patch("configparser_override.file_collector._unix_collect_home_config")
def test_config_file_collector_darwin(
    mock_home_config, mock_system_config, mock_platform
):
    mock_system_config.return_value = [Path("/etc/xdg/testapp/config.ini")]
    mock_home_config.return_value = Path("/home/testuser/.config/testapp/config.ini")
    result = config_file_collector("config.ini", "testapp")
    assert result == [
        Path("/etc/xdg/testapp/config.ini"),
        Path("/home/testuser/.config/testapp/config.ini"),
    ]


@patch("configparser_override.file_collector.platform.system", return_value="Linux")
@patch("configparser_override.file_collector._unix_collect_system_config")
@patch("configparser_override.file_collector._unix_collect_home_config")
def test_config_file_collector_linux_only_1_file(
    mock_home_config, mock_system_config, mock_platform
):
    mock_system_config.return_value = [
        Path("/etc/xdg/testapp/config.ini"),
    ]
    mock_home_config.return_value = Path("/home/testuser/.config/testapp/config.ini")
    result = config_file_collector(
        "config.ini", "testapp", only_most_important_file=True
    )
    assert result == [
        Path("/home/testuser/.config/testapp/config.ini"),
    ]


@patch("configparser_override.file_collector.platform.system", return_value="Linux")
@patch("configparser_override.file_collector._unix_collect_system_config")
@patch("configparser_override.file_collector._unix_collect_home_config")
def test_config_file_collector_linux_multiple_system_files(
    mock_home_config, mock_system_config, mock_platform
):
    mock_system_config.return_value = [
        Path("/etc/xdg/testapp/config.ini"),
        Path("/etc/testapp/config.ini"),
    ]
    mock_home_config.return_value = Path("/home/testuser/.config/testapp/config.ini")
    result = config_file_collector("config.ini", "testapp")
    assert result == [
        Path("/etc/xdg/testapp/config.ini"),
        Path("/etc/testapp/config.ini"),
        Path("/home/testuser/.config/testapp/config.ini"),
    ]


@patch("configparser_override.file_collector.platform.system", return_value="Linux")
@patch("configparser_override.file_collector._unix_collect_system_config")
@patch("configparser_override.file_collector._unix_collect_home_config")
def test_config_file_collector_no_files_found(
    mock_home_config, mock_system_config, mock_platform
):
    mock_system_config.return_value = []
    mock_home_config.return_value = None
    with pytest.raises(NoConfigFilesFoundError):
        config_file_collector("config.ini", "testapp", allow_no_found_files=False)



import logging
import os
import platform
from pathlib import Path
from typing import List, Optional

from configparser_override.exceptions import NoConfigFilesFoundError

logger = logging.getLogger(__name__)


def _log_and_return_if_exists(file_path: Path) -> Optional[Path]:
    """
    Check if the given file path exists and log as found if so.

    :param file_path: Path to the configuration file.
    :type file_path: Path
    :return: The file path if it exists, otherwise None.
    :rtype: Optional[Path]
    """
    if file_path.exists():
        logger.debug(f"Found config file: {file_path}")
        return file_path
    return None


def _unix_collect_home_config(subdir: str, file_name: str) -> Optional[Path]:
    """
    Collect Unix home directory configuration file.

    :param subdir: Subdirectory under the home config path.
    :type subdir: str
    :param file_name: Name of the configuration file.
    :type file_name: str
    :return: Path to the home configuration file if it exists, otherwise None.
    :rtype: Optional[Path]
    """
    home = Path.home()
    xdg_config_home = Path(os.getenv("XDG_CONFIG_HOME", home / ".config"))
    home_config = xdg_config_home / subdir / file_name
    return _log_and_return_if_exists(home_config)


def _unix_collect_system_config(
    subdir: str, file_name: str, bare_etc: bool = False
) -> List[Path]:
    """
    Collect Unix system configuration files.

    :param subdir: Subdirectory under the system config path.
    :type subdir: str
    :param file_name: Name of the configuration file.
    :type file_name: str
    :param bare_etc: Look in bare `/etc` directory if True.
    :type bare_etc: bool
    :return: List of paths to system configuration files.
    :rtype: List[Path]
    """
    config_file_list = []
    if bare_etc:
        file_path = Path("/etc") / subdir / file_name
        config_file = _log_and_return_if_exists(file_path)
        return [config_file] if config_file else []

    xdg_config_dirs = [
        Path(dir) for dir in os.getenv("XDG_CONFIG_DIRS", "/etc/xdg").split(":")
    ]
    for dir in xdg_config_dirs:
        file_path = dir / subdir / file_name
        config_file = _log_and_return_if_exists(file_path)
        if config_file:
            config_file_list.append(config_file)
    config_file_list.reverse()
    return config_file_list


def _windows_collect_home_config(subdir: str, file_name: str) -> Optional[Path]:
    """
    Collect Windows home directory configuration file.

    :param subdir: Subdirectory under the home config path.
    :type subdir: str
    :param file_name: Name of the configuration file.
    :type file_name: str
    :return: Path to the home configuration file if it exists, otherwise None.
    :rtype: Optional[Path]
    """
    appdata = os.getenv("APPDATA")
    if appdata:
        home_config = Path(appdata) / subdir / file_name
        return _log_and_return_if_exists(home_config)
    return None


def _windows_collect_system_config(subdir: str, file_name: str) -> List[Path]:
    """
    Collect Windows system configuration files.

    :param subdir: Subdirectory under the system config path.
    :type subdir: str
    :param file_name: Name of the configuration file.
    :type file_name: str
    :return: List of paths to system configuration files.
    :rtype: List[Path]
    """
    programdata = os.getenv("PROGRAMDATA")
    config_file_list = []
    if programdata:
        file_path = Path(programdata) / subdir / file_name
        config_file = _log_and_return_if_exists(file_path)
        if config_file:
            config_file_list.append(config_file)
    return config_file_list


def config_file_collector(
    file_name: str,
    app_name: str = "",
    only_most_important_file: bool = False,
    allow_no_found_files: bool = True,
    bare_etc: bool = False,
) -> List[Path]:
    """
    Collect configuration files from conventional locations.

    If multiple files are returned, they are in order of ascending priority
    (most important file last).

    :param file_name: Name of the configuration file.
    :type file_name: str
    :param app_name: Name of the app, used as a subdirectory.
    :type app_name: str
    :param only_most_important_file: Only return the most important (prioritized) file,
        eg. home directory config files is more important than system wide config file
    :type only_most_important_file: bool
    :param allow_no_found_files: Whether to allow no found files without raising an error.
    :type allow_no_found_files: bool
    :param bare_etc: Look in bare `/etc` directory if True (Unix only).
    :type bare_etc: bool
    :return: List of paths to configuration files.
    :rtype: List[Path]
    :raises NoConfigFilesFoundError: If no config files are found and `allow_no_found_files` is False.

    .. code-block:: python

        from pathlib import Path
        from your_module import config_file_collector

        # Collect configuration files for the my_app application with
        # the name config.ini
        config_files = config_file_collector(
            file_name="config.ini",
            app_name="my_app",
            only_most_important_file=False,
            allow_no_found_files=True,
        )

        # Output the found configuration files
        for config_file in config_files:
            print(f"Found config file: {config_file}")
        # Example output:
        #    Found config file: /etc/xdg/my_app/config.ini
        #    Found config file: /home/user/.config/my_app/config.ini

    """
    system = platform.system()

    if system == "Windows":
        config_files = _windows_collect_system_config(app_name, file_name)
        home_config = _windows_collect_home_config(app_name, file_name)
    else:
        config_files = _unix_collect_system_config(app_name, file_name, bare_etc)
        home_config = _unix_collect_home_config(app_name, file_name)

    if home_config:
        config_files.append(home_config)

    if not config_files and not allow_no_found_files:
        raise NoConfigFilesFoundError(
            f"No configuration files found for file_name={file_name}, app_name={app_name}"
        )

    return (
        [config_files.pop()]
        if config_files and only_most_important_file
        else config_files
    )



import logging

logger = logging.getLogger(__name__)


class ConfigParserOverrideError(Exception):
    pass


class OverrideStrategyNotImplementedError(ConfigParserOverrideError):
    """Exception raised when an unimplemented strategy is requested."""

    pass


class SectionNotFound(ConfigParserOverrideError):
    """Exception raised when a section is not found in the ConfigParser"""

    pass


class ConversionError(ConfigParserOverrideError):
    """Exception raised for errors of casting a string to typehinted value(s)"""

    pass


class LiteralEvalMiscast(ConfigParserOverrideError):
    """
    Exception raised when ast.literal_eval casts the string to a unexpected data type
    """

    pass


class NoConfigFilesFoundError(ConfigParserOverrideError):
    """
    Exception raised when no configuration files can be found
    """

    pass


class InvalidParametersError(ConfigParserOverrideError):
    """
    Exception raised when invalid parameters are used
    """

    pass


class ConversionIgnoreError(ConfigParserOverrideError):
    """
    Exeption raised when a dataclass field can not be skipped during
    conversion. Because the field is not Optional nor have a default or
    default_factory assignment.
    """

    pass



import configparser
import platform
from dataclasses import dataclass, fields
from pathlib import Path
from typing import Any, Callable, Optional, get_type_hints

import pytest

from configparser_override import ConfigParserOverride
from configparser_override.convert import (
    ConfigConverter,
    _can_ignore_conversion,
    _can_ignore_section,
    _field_has_default_value,
    _is_optional_dataclass,
    _is_optional_type,
)
from configparser_override.exceptions import (
    ConversionError,
    ConversionIgnoreError,
    InvalidParametersError,
    LiteralEvalMiscast,
)


@pytest.fixture()
def config_file_simple_types(tmp_path):
    config_content = """
    [DEFAULT]
    allkey = string

    [section1]
    key1 = 123
    key2 = string

    [section2]
    key3 = 1.2
    """
    config_path = tmp_path / "config.ini"
    config_path.write_text(config_content)
    return str(config_path)


@dataclass
class DefaultSection:
    allkey: str


@dataclass
class Section1:
    key1: int
    key2: str
    allkey: str


@dataclass
class Section2:
    key3: float
    allkey: str


@dataclass
class ConfigFileSimpleTypes:
    DEFAULT: DefaultSection
    section1: Section1
    section2: Section2


@pytest.fixture()
def config_file_complex_types(tmp_path):
    config_content = """
    [DEFAULT]
    allkey = byte
    optionalallkey = 123

    [section1]
    key1 = [1,2,3,4,5,6]
    key2 = {"key": "value", "key2": "value2"}

    [section2]
    key3 = [1.2,1.4]
    """
    config_path = tmp_path / "config.ini"
    config_path.write_text(config_content)
    return str(config_path)


@dataclass
class ComplexDefaultSection:
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexSection1:
    key1: list[int | str]
    key2: dict[str, str]
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexSection2:
    key3: list[float]
    allkey: bytes
    optionalallkey: Optional[str | int]
    key4: Optional[float] = None


@dataclass
class ConfigFileComlexTypes:
    DEFAULT: ComplexDefaultSection
    section1: ComplexSection1
    section2: ComplexSection2


@pytest.fixture()
def config_file_complex_types_nested(tmp_path):
    config_content = """
    [DEFAULT]
    allkey = byte
    optionalallkey = 123

    [section1]
    key1 = [[1,2,3],[4,5,6]]
    key2 = {"key": {"nestedkey": "value"}, "key2": {"nestedkey2": "value2"}}

    [section2]
    key3 = ["true",1,"yes","no",0,"false"]

    [section3]
    key4 = {"string1","string2"}
    key5 = [{"string1"},{"string2"}]
    key6 = ({"123j"},{"4+2j"})
    """
    config_path = tmp_path / "config.ini"
    config_path.write_text(config_content)
    return str(config_path)


@dataclass
class ComplexNestedDefaultSection:
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexNestedSection1:
    key1: list[list[int | str]]
    key2: dict[str, dict[str, str]]
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexNestedSection2:
    key3: list[Optional[bool | str]]
    allkey: bytes
    optionalallkey: Optional[str | int]
    key4: Optional[float] = None


@dataclass
class ComplexNestedSection3:
    key4: set[str]
    key5: list[set[str]]
    key6: tuple[set[complex]]


@dataclass
class ConfigFileComlexNestedTypes:
    DEFAULT: ComplexNestedDefaultSection
    section1: ComplexNestedSection1
    section2: ComplexNestedSection2
    section3: ComplexNestedSection3


def test_simple_config_to_dataclass(config_file_simple_types):
    parser = ConfigParserOverride()
    parser.read(filenames=config_file_simple_types)
    parser.apply_overrides()

    dataclass_rep = ConfigConverter(parser.config).to_dataclass(ConfigFileSimpleTypes)
    assert dataclass_rep.DEFAULT.allkey == "string"
    assert dataclass_rep.section1.allkey == "string"
    assert dataclass_rep.section1.key1 == 123
    assert dataclass_rep.section1.key2 == "string"
    assert dataclass_rep.section2.key3 == 1.2


def test_complex_config_to_dataclass(config_file_complex_types):
    parser = ConfigParserOverride()
    parser.read(filenames=config_file_complex_types)
    parser.apply_overrides()

    dataclass_rep = ConfigConverter(parser.config).to_dataclass(ConfigFileComlexTypes)
    assert dataclass_rep.DEFAULT.allkey == b"byte"
    assert dataclass_rep.section1.allkey == b"byte"
    assert dataclass_rep.section1.key1 == [1, 2, 3, 4, 5, 6]
    assert dataclass_rep.section1.key2 == {
        "key": "value",
        "key2": "value2",
    }
    assert dataclass_rep.section2.key3 == [1.2, 1.4]


def test_complex_nested_config_to_dataclass(config_file_complex_types_nested):
    parser = ConfigParserOverride()
    parser.read(filenames=config_file_complex_types_nested)
    parser.apply_overrides()

    dataclass_rep = ConfigConverter(parser.config).to_dataclass(
        ConfigFileComlexNestedTypes
    )
    assert dataclass_rep.DEFAULT.allkey == b"byte"
    assert dataclass_rep.section1.allkey == b"byte"
    assert dataclass_rep.section1.key1 == [[1, 2, 3], [4, 5, 6]]
    assert dataclass_rep.section1.key2 == {
        "key": {"nestedkey": "value"},
        "key2": {"nestedkey2": "value2"},
    }
    assert dataclass_rep.section2.key3 == [True, True, True, False, False, False]
    assert dataclass_rep.section3.key4 == {"string1", "string2"}
    assert dataclass_rep.section3.key5 == [{"string1"}, {"string2"}]
    assert dataclass_rep.section3.key6 == ({123j}, {4 + 2j})


def test_config_to_dataclass_custom_bools():
    custom_booleans = {"cool": True, "not cool": False}

    @dataclass
    class Sect1:
        true: bool
        false: bool

    @dataclass
    class CustomBools:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__true="cool", sect1__false="not cool")
    parser.read(filenames=[])
    parser.apply_overrides()

    dataclass_rep = ConfigConverter(
        parser.config, boolean_states=custom_booleans
    ).to_dataclass(CustomBools)

    assert dataclass_rep.sect1.false is False
    assert dataclass_rep.sect1.true is True


def test_config_to_dataclass_bools_not_valid():
    @dataclass
    class Sect1:
        true: bool
        false: bool

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(
        sect1__true="notValidTrue", sect1__false="notValidFalse"
    )
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ValueError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_missing_key_in_config():
    @dataclass
    class Sect1:
        key: str
        key132: str

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key="ok")
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ConversionIgnoreError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_ignore_section_config():
    @dataclass
    class Sect1:
        key: str

    @dataclass
    class Sect2:
        key: str
        key132: str

    @dataclass
    class C:
        sect1: Sect1
        sect2: Optional[Sect2] = None

    parser = ConfigParserOverride(sect1__key="ok", sect2__key="ignore_me")
    parser.read(filenames=[])
    parser.apply_overrides()

    dc_config = ConfigConverter(parser.config).to_dataclass(C)
    assert parser.config["sect2"]["key"] == "ignore_me"
    assert dc_config.sect1.key == "ok"
    assert dc_config.sect2 is None


def test_dont_ignore_default_none_section_config():
    @dataclass
    class Sect1:
        key: str

    @dataclass
    class Sect2:
        key: str
        key123: str

    @dataclass
    class C:
        sect1: Sect1
        sect2: Optional[Sect2] = None

    config = configparser.ConfigParser()
    config.add_section("sect1")
    config.set(section="sect1", option="key", value="a")
    config.add_section("sect2")
    config.set(section="sect2", option="key", value="b")
    config.set(section="sect2", option="key123", value="c")

    dc_config = ConfigConverter(config).to_dataclass(C)
    assert config["sect1"]["key"] == "a"
    assert config["sect2"]["key"] == "b"
    assert dc_config.sect1.key == "a"
    assert dc_config.sect2 is not None
    assert dc_config.sect2.key == "b"
    assert dc_config.sect2.key123 == "c"


def test_optional_section_config():
    @dataclass
    class Sect1:
        key: str

    @dataclass
    class Sect2:
        key: str
        key123: str

    @dataclass
    class C:
        sect1: Sect1
        sect2: Optional[Sect2] = None

    parser = ConfigParserOverride(
        sect1__key="ok", sect2__key="ok1", sect2__key123="ok123"
    )
    parser.read(filenames=[])
    parser.apply_overrides()

    dc_config = ConfigConverter(parser.config).to_dataclass(C)
    assert parser.config["sect2"]["key"] == "ok1"
    assert dc_config.sect1.key == "ok"
    assert dc_config.sect2 is not None
    assert dc_config.sect2.key == "ok1"


def test_exclude_section_config():
    @dataclass
    class Sect1:
        key: str

    @dataclass
    class Sect2:
        key: str
        key123: str

    @dataclass
    class C:
        sect1: Sect1
        sect2: Optional[Sect2] = None

    parser = ConfigParserOverride(
        sect1__key="ok",
    )
    parser.read(filenames=[])
    parser.apply_overrides()

    dc_config = ConfigConverter(parser.config, exclude_sections=["sect2"]).to_dataclass(
        C
    )
    assert dc_config.sect1.key == "ok"
    assert dc_config.sect2 is None


def test_exclude_non_optional_section_config():
    @dataclass
    class Sect1:
        key: str

    @dataclass
    class Sect2:
        key: str
        key123: str

    @dataclass
    class C:
        sect1: Sect1
        sect2: Sect2

    parser = ConfigParserOverride(
        sect1__key="ok",
    )
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ConversionIgnoreError):
        ConfigConverter(parser.config, exclude_sections=["sect2"]).to_dataclass(C)


def test_exclude_optional_section_config():
    @dataclass
    class Sect1:
        key: str

    @dataclass
    class Sect2:
        key: str
        key123: str

    @dataclass
    class C:
        sect1: Sect1
        sect2: Optional[Sect2] = None

    parser = ConfigParserOverride(
        sect1__key="ok",
    )
    parser.read(filenames=[])
    parser.apply_overrides()

    dc_config = ConfigConverter(parser.config, exclude_sections=["sect2"]).to_dataclass(
        C
    )
    assert dc_config.sect2 is None
    assert dc_config.sect1.key == "ok"


def test_unsupported_type():
    @dataclass
    class Sect1:
        key: Callable

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key="ok")
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ValueError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_list_member_conversion_error():
    @dataclass
    class Sect1:
        key: list[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_set_member_conversion_error():
    @dataclass
    class Sect1:
        key: set[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='{"ok"}')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_dict_member_conversion_error():
    @dataclass
    class Sect1:
        key: dict[str, int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='{"ok":"ok"}')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_tuple_member_conversion_error():
    @dataclass
    class Sect1:
        key: tuple[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='("ok","ok")')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_union_conversion_error():
    @dataclass
    class Sect1:
        key: list[int | float]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok","ok"]')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).to_dataclass(C)


def test_list_member_literaleval_error():
    @dataclass
    class Sect1:
        key: list[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='{"ok"}')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).to_dataclass(C)


def test_set_member_literaleval_error():
    @dataclass
    class Sect1:
        key: set[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).to_dataclass(C)


def test_dict_member_literaleval_error():
    @dataclass
    class Sect1:
        key: dict[str, int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).to_dataclass(C)


def test_tuple_member_literaleval_error():
    @dataclass
    class Sect1:
        key: tuple[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])
    parser.apply_overrides()

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).to_dataclass(C)


@pytest.fixture()
def config_file_allow_empty(tmp_path):
    config_content = """
    [sectionv]
    key1 = 123

    [sectione]
    key3
    """
    config_path = tmp_path / "config_allow_empty.ini"
    config_path.write_text(config_content)
    return str(config_path)


def test_none_type(config_file_allow_empty):
    @dataclass
    class SectionV:
        key1: int

    @dataclass
    class SectionE:
        key3: None

    @dataclass
    class ConfigEmptyKey:
        sectionv: SectionV
        sectione: SectionE

    parser = ConfigParserOverride(
        config_parser=configparser.ConfigParser(allow_no_value=True)
    )
    parser.read(filenames=[config_file_allow_empty])
    parser.apply_overrides()

    dataclass_rep = parser.to_dataclass(ConfigEmptyKey)
    assert dataclass_rep.sectione.key3 is None
    assert dataclass_rep.sectionv.key1 == 123


@pytest.fixture()
def config_file_any(tmp_path):
    config_content = """
    [section]
    key1 = 123
    """
    config_path = tmp_path / "config_any.ini"
    config_path.write_text(config_content)
    return str(config_path)


def test_any_type(config_file_any):
    @dataclass
    class Section:
        key1: Any

    @dataclass
    class ConfigAny:
        section: Section

    parser = ConfigParserOverride(config_parser=configparser.ConfigParser())
    parser.read(filenames=[config_file_any])
    parser.apply_overrides()

    dataclass_rep = parser.to_dataclass(ConfigAny)
    assert dataclass_rep.section.key1 == "123"


@pytest.fixture()
def config_file_path(tmp_path):
    config_content = """
    [section]
    key1 = relative/path
    key2 = /absolut/unix/path
    key3 = c:/absolut/windows/path
    """
    config_path = tmp_path / "config_any.ini"
    config_path.write_text(config_content)
    return str(config_path)


def test_path_type(config_file_path):
    @dataclass
    class Section:
        key1: Path
        key2: Path
        key3: Path

    @dataclass
    class ConfigAny:
        section: Section

    parser = ConfigParserOverride(config_parser=configparser.ConfigParser())
    parser.read(filenames=[config_file_path])
    parser.apply_overrides()

    dataclass_rep = parser.to_dataclass(ConfigAny)
    assert dataclass_rep.section.key1 == Path("relative/path")
    assert dataclass_rep.section.key2 == Path("/absolut/unix/path")
    assert dataclass_rep.section.key3 == Path("c:/absolut/windows/path")
    assert dataclass_rep.section.key1.resolve().is_relative_to(Path.cwd())
    assert not dataclass_rep.section.key1.is_absolute()

    system = platform.system()

    if system == "Windows":
        assert not dataclass_rep.section.key2.is_absolute()
        assert dataclass_rep.section.key3.is_absolute()
    else:  # UNIX
        assert dataclass_rep.section.key2.is_absolute()
        assert not dataclass_rep.section.key3.is_absolute()


def test_is_optional_type_true():
    assert _is_optional_type(Optional[int])
    assert _is_optional_type(Optional[str])


def test_is_optional_type_false():
    assert not _is_optional_type(int)
    assert not _is_optional_type(str)


def test_is_optional_dataclass_true():
    @dataclass
    class C:
        a: int

    @dataclass
    class C2:
        a: int
        b: C
        c: Optional[C] = None

    t = get_type_hints(C2)
    assert _is_optional_dataclass(Optional[C])
    assert _is_optional_dataclass(t["c"])
    assert not _is_optional_dataclass(t["b"])


def test_is_optional_dataclass_false():
    @dataclass
    class C:
        a: int

    assert not _is_optional_dataclass(C)
    assert not _is_optional_dataclass(C(a=1))
    assert not _is_optional_dataclass(Optional[int])


def test_can_ignore_section():
    @dataclass
    class C:
        a: int

    @dataclass
    class C2:
        c: Optional[C] = None

    f = fields(C2)
    fc = fields(C)
    assert _can_ignore_section(f[0])
    assert not _can_ignore_section(fc[0])


def test_can_ignore_conversion():
    @dataclass
    class C:
        a: int

    @dataclass
    class C2:
        c: Optional[C] = None

    f = fields(C2)
    fc = fields(C)
    assert _can_ignore_conversion(f[0])
    assert not _can_ignore_conversion(fc[0])


def test_field_has_default_value():
    @dataclass
    class C1:
        a: int = 1

    @dataclass
    class C:
        a: int

    @dataclass
    class C2:
        c: Optional[C] = None

    f = fields(C)
    f1 = fields(C1)
    f2 = fields(C2)
    assert not _field_has_default_value(f[0])
    assert _field_has_default_value(f1[0])
    assert _field_has_default_value(f2[0])


def test_parse_section():
    config = configparser.ConfigParser()
    config.add_section("abc")
    config.set(section="abc", option="a", value="1")
    config.add_section("def")
    config.set(section="def", option="a", value="1")

    converter = ConfigConverter(config, include_sections=["abc"])
    assert converter._parse_section("abc")
    assert not converter._parse_section("def")


def test_too_many_args():
    config = configparser.ConfigParser()

    with pytest.raises(InvalidParametersError):
        ConfigConverter(config, include_sections=["abc"], exclude_sections=["def"])



from __future__ import annotations

import ast
import dataclasses
import logging
from pathlib import Path
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

if TYPE_CHECKING:
    import configparser

    from configparser_override.types import Dataclass

from configparser_override.exceptions import (
    ConversionError,
    ConversionIgnoreError,
    InvalidParametersError,
    LiteralEvalMiscast,
)

logger = logging.getLogger(__name__)


def _is_optional_type(type_hint: Any) -> bool:
    """
    Check if a given type hint is an optional type.
    """
    return get_origin(type_hint) in [Union, UnionType] and type(None) in get_args(
        type_hint
    )


def _is_optional_dataclass(type_hint: Any) -> bool:
    """
    Check if a given type hint is an optional dataclass.
    """
    if get_origin(type_hint) not in [Union, UnionType]:
        return False

    for arg in get_args(type_hint):
        if arg is type(None) or dataclasses.is_dataclass(arg):
            continue
        else:
            return False

    return True


def _field_has_default_value(field: dataclasses.Field) -> bool:
    """
    Check if a given dataclass field has a default value.
    """
    return (
        field.default_factory != dataclasses.MISSING
        or field.default != dataclasses.MISSING
    )


def _can_ignore_section(field: dataclasses.Field) -> bool:
    return _is_optional_dataclass(field.type) or _field_has_default_value(field)


def _can_ignore_conversion(field: dataclasses.Field) -> bool:
    return _is_optional_type(field.type) or _field_has_default_value(field)


class ConfigConverter:
    """
    A class to convert configuration data from a ConfigParser object to a dictionary
    or dataclass.

    :param config: The configuration parser object.
    :type config: configparser.ConfigParser
    :param boolean_states: Optional mapping of custom boolean states,
        defaults to None and uses the internal mapping of the ConfigParser object.
    :type boolean_states: Optional[Mapping[str, bool]], optional
    """

    def __init__(
        self,
        config: configparser.ConfigParser,
        boolean_states: Optional[Mapping[str, bool]] = None,
        include_sections: Optional[List[str]] = None,
        exclude_sections: Optional[List[str]] = None,
    ) -> None:
        self.config = config

        if include_sections is not None and exclude_sections is not None:
            raise InvalidParametersError(
                "Not allows to specify both include_sections and exclude_sections parameters at the same time"
            )
        self.include_sections = include_sections
        self.exclude_sections = exclude_sections

        if boolean_states:
            self.boolean_states = boolean_states
        else:
            self.boolean_states = self.config.BOOLEAN_STATES

    def to_dataclass(self, dataclass: Type[Dataclass]) -> Dataclass:
        """
        Convert the configuration data to a dataclass instance.

        :param dataclass: The dataclass type to convert the configuration data into.
        :type dataclass: Dataclass
        :return: An instance of the dataclass populated with the configuration data.
        :rtype: Dataclass

        **Examples:**

        .. code-block:: python

            >>> from dataclasses import dataclass

            >>> @dataclass
            ... class Section1:
            ...     key: str

            >>> @dataclass
            ... class ExampleConfig:
            ...     section1: Section1

            >>> config = configparser.ConfigParser()
            >>> config.read_string(\"\"\"
            ... [section1]
            ... key = value
            ... \"\"\")
            >>> converter = ConfigConverter(config)
            >>> config_as_dataclass = converter.to_dataclass(ExampleConfig)
            >>> assert config_as_dataclass.section1.key == "value" # True
        """
        config_dict = self._to_dict()
        return self._dict_to_dataclass(
            input_dict=config_dict,
            dataclass=dataclass,
        )

    def _to_dict(self) -> dict[str, dict[str, str]]:
        """
        Convert the configuration data to a nested dictionary.

        :return: The configuration data as a dictionary.
        :rtype: dict[str, dict[str, str]]

        **Examples:**

        .. code-block:: python

            >>> config = configparser.ConfigParser()
            >>> config.read_string(\"\"\"
            ... [section1]
            ... key1 = value1
            ... [section2]
            ... key2 = value2
            ... \"\"\")
            >>> converter = ConfigConverter(config)
            >>> config_dict = converter._to_dict()
            >>> config_dict['section1']['key1']
            'value1'
        """
        config_dict: dict[str, dict[str, str]] = {}
        for sect in self.config.sections():
            # Add nested sections
            config_dict[sect] = {}
            for opt in self.config.options(sect):
                config_dict[sect][opt] = self.config.get(section=sect, option=opt)
        # Add default nested section
        config_dict[self.config.default_section] = {}
        for opt in self.config.defaults():
            config_dict[self.config.default_section][opt] = self.config.get(
                section=self.config.default_section, option=opt
            )
        return config_dict

    def _dict_to_dataclass(
        self,
        input_dict: dict,
        dataclass: Type[Dataclass],
        nested_level: int = 0,
    ) -> Dataclass:
        type_hints = get_type_hints(dataclass)

        _dict_with_types: dict[str, Any] = {}
        for field in dataclasses.fields(dataclass):
            field_name = field.name
            field_type = type_hints[field_name]
            logger.debug(f"Initiate conversion of {field_name=} and {field_type=}")

            # Skip convertion of specified sections
            if nested_level == 0 and not self._parse_section(field_name):
                if _can_ignore_section(field):
                    logger.debug(f"Ignore conversion of section {field_name}")
                    continue
                else:
                    raise ConversionIgnoreError(
                        f"Can not skip {field_name=}, the field is not optional nor have a default or default_factory assignment."
                    )

            # Create dict with field names and casted values
            if field_name in input_dict:
                logger.debug(f"Initiate type cast of {field_name=} to {field_type=}")
                _dict_with_types[field_name] = self._cast_value(
                    value=input_dict[field_name],
                    type_hint=field_type,
                    nested_level=nested_level,
                )
            elif not _can_ignore_conversion(field):
                raise ConversionIgnoreError(
                    f"Config not found and not allowed to skip {field_name=}, the field is not optional nor have a default or default_factory assignment."
                )
        return dataclass(**_dict_with_types)

    def _parse_section(self, section: str) -> bool:
        is_not_included = (
            self.include_sections is not None and section not in self.include_sections
        )
        is_excluded = (
            self.exclude_sections is not None and section in self.exclude_sections
        )

        return not (is_not_included or is_excluded)

    def _cast_value(self, value: Any, type_hint: Any, nested_level: int = 0) -> Any:
        if dataclasses.is_dataclass(type_hint):
            logger.debug("Type hint is a Dataclass")
            _type_hint = type_hint if isinstance(type_hint, type) else type(type_hint)
            return self._dict_to_dataclass(
                input_dict=value, dataclass=_type_hint, nested_level=nested_level + 1
            )
        if type_hint is Any:
            return value
        if type_hint in [int, float, complex, str, Path]:
            return type_hint(value)
        if type_hint is bytes:
            return str(value).encode()
        if type_hint is bool:
            return self._cast_bool(value)
        _origin = get_origin(type_hint)
        if _origin in [list, List]:
            return self._cast_list(value, type_hint)
        if _origin in [dict, Dict]:
            return self._cast_dict(value, type_hint)
        if _origin in [set, Set]:
            return self._cast_set(value, type_hint)
        if _origin in [tuple, Tuple]:
            return self._cast_tuple(value, type_hint)
        if _origin in (Optional, Union, UnionType):
            return self._cast_union(value, type_hint)
        if type_hint is type(None):
            return None
        raise ValueError(f"Unsupported type: {type_hint}")

    def _cast_bool(self, value: Any) -> bool:
        if str(value).lower() in self.boolean_states:
            return self.boolean_states[str(value).lower()]
        else:
            raise ValueError(f"{value=} not in possible {self.boolean_states=}")

    def _cast_list(self, value: Any, type_hint: Any) -> list:
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, list):
            _types = get_args(type_hint)
            for typ in _types:
                try:
                    return [self._cast_value(item, typ) for item in _evaluated_option]
                except Exception as e:
                    logger.debug(f"Failed to cast {value=} into {typ=}, error: {e}")
                    continue
            raise ConversionError(
                f"Not possible to cast {value} into a list of {_types}"
            )
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_set(self, value: Any, type_hint: Any) -> set:
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, set):
            _types = get_args(type_hint)
            for typ in _types:
                try:
                    return {self._cast_value(item, typ) for item in _evaluated_option}
                except Exception as e:
                    logger.debug(f"Failed to cast {value=} into {typ=}, error: {e}")
                    continue
            raise ConversionError(
                f"Not possible to cast {value} into a set of {_types}"
            )
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_tuple(self, value: Any, type_hint: Any) -> tuple:
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, tuple):
            _types = get_args(type_hint)
            for typ in _types:
                try:
                    return tuple(
                        self._cast_value(item, typ) for item in _evaluated_option
                    )
                except Exception as e:
                    logger.debug(f"Failed to cast {value=} into {typ=}, error: {e}")
                    continue
            raise ConversionError(
                f"Not possible to cast {value} into a tuple of {_types}"
            )
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_dict(self, value: Any, type_hint: Any) -> dict:
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, dict):
            k_typ, v_typ = get_args(type_hint)
            try:
                return {
                    self._cast_value(k, k_typ): self._cast_value(v, v_typ)
                    for k, v in _evaluated_option.items()
                }
            except Exception as e:
                logger.debug(
                    f"Failed to cast {value=} into {k_typ=}, {v_typ=}, error: {e}"
                )
                raise ConversionError(
                    f"Not possible to cast {value} into a dict of keys of type {k_typ}, and values of type {v_typ}, Error: {e}"
                ) from e
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_union(self, value: Any, type_hint: Any) -> Any:
        for typ in get_args(type_hint):
            try:
                return self._cast_value(value, typ)
            except Exception as e:
                logger.debug(f"Failed to cast {value=} into {typ=}, error: {e}")
                continue
        raise ConversionError(f"Not possible to cast {value} into type {type_hint}")
